
#import "NamingContext.h"
#import <ADORB/ADORB.h>

@implementation NamingContext

static NSString* _name_separator = @"\n";
static NSString* interfacePath = @"CosNaming/NamingContext";

- interface
{
    return _interface;
}

- (void) invalidate:(NSNotification*)notification
{
    id object = [notification object];
    NSEnumerator* en = [_entries keyEnumerator];
    NSString* key;
    while(key = [en nextObject])
        if([_entries objectForKey:key] == object)
            [_entries removeObjectForKey:key];
}

- init
{
    self = [super init];
    _entries = [[NSMutableDictionary allocWithZone:[self zone]]init];
    _interface = [[ORB defaultORB]interfaceNamed:interfacePath];
    [[NSNotificationCenter defaultCenter]addObserver:self selector:@selector(invalidate:)
        name:VXInvalidateObject object:nil];
    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_entries release];
    [super dealloc];
}

- (BOOL) isNamingContext:obj
{
    if([obj respondsToSelector:@selector(interface)]){
        IDLInterface* objInterface = [obj interface];
        if([objInterface isKindOfClass:[IDLInterface class]])
            return [objInterface adoptsInterface:_interface];
    }
    return NO;
}

- (NSString*) stringFromName:(NSDictionary*)name
{
    // Use \n to separate 'id' and 'kind'. Assuming 'id' is not nil.
    NSString* kind = [name valueForKey:@"kind"];
    return [NSString stringWithFormat:@"%@%@%@",
        [name valueForKey:@"id"],
        _name_separator,
        kind?kind:@""];
}

- (NSDictionary*) nameComponentFromString:(NSString*)str
{
    NSRange r = [str rangeOfString:_name_separator];
    NSString* _id;
    NSString* _kind = @"";
    if(r.length > 0){
        _id = [str substringToIndex:r.location];
        _kind = [str substringFromIndex:r.location + r.length];
    }else
        _id = str;
    return [NSDictionary dictionaryWithObjectsAndKeys:
        _id, @"id", _kind, @"kind", nil];
}

- (NSArray*) nameFromString:(NSString*)str
{
    return [NSArray arrayWithObject:[self nameComponentFromString:str]];
}

- (void) doBind:(SEL)sel name:(NSArray*)nameArray object:obj
{
    int cnt = [nameArray count];
    if(cnt > 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id ctx = [_entries objectForKey:key];
        if(ctx && [self isNamingContext:ctx]){
            NSRange r = { 1, cnt-1};
            NSArray* rest = [nameArray subarrayWithRange:r];
            [ctx performSelector:sel withObject:rest withObject:obj];
        }else{
            NSString* reasonKey = ctx?@"not_context":@"missing_node";
            CORBAException* excp = [_interface exceptionWithName:@"NotFound"];
            id reason = [_interface enumValueForName:@"NotFoundReason" andKey:reasonKey];
            [excp takeValue:reason forKey:@"why"];
            [excp takeValue:nameArray forKey:@"rest_of_name"];
            [excp raise];
        }
    }else
        [NSException raise:@"CORBAInternalException"
            format:@"Internal inconsistency"];

}

- (void) bind:nameArray object:obj
{
    int cnt = [nameArray count];
    if(cnt > 1){
        [self doBind:_cmd name:nameArray object:obj];
    }else if(cnt == 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        if(![_entries objectForKey:key])
            [_entries setObject:obj forKey:key];
        else{
            CORBAException* excp = [_interface exceptionWithName:@"AlreadyBound"];
            [excp raise];
        }
    }else{
        CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
        [excp raise];
    }
}

- (void) rebind:nameArray object:obj
{
    int cnt = [nameArray count];
    if(cnt > 1){
        [self doBind:_cmd name:nameArray object:obj];
    }else if(cnt == 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        [_entries setObject:obj forKey:key];
    }else{
        CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
        [excp raise];
    }
}

- (void) bindContext:nameArray namingContext:obj
{
    int cnt = [nameArray count];
    if(cnt > 1){
        [self doBind:_cmd name:nameArray object:obj];
    }else if(cnt == 1){
        if([self isNamingContext:obj])
            [self bind:nameArray object:obj];
        else{
            // No specific exception for this case
            CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
            [excp raise];
        }
    }else{
        CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
        [excp raise];
    }
}

- (void) rebindContext:nameArray namingContext:obj
{
    int cnt = [nameArray count];
    if(cnt > 1){
        [self doBind:_cmd name:nameArray object:obj];
    }else if(cnt == 1){
        if([self isNamingContext:obj])
            [self rebind:nameArray object:obj];
        else{
            // No specific exception for this case
            CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
            [excp raise];
        }
    }else{
        CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
        [excp raise];
    }
}

- resolve:nameArray
{
    int cnt = [nameArray count];
    if(cnt > 0){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id obj = [_entries objectForKey:key];
        if(obj){
            if(cnt == 1)
                return obj;
            else{
                NSRange r = {1,cnt-1};
                NSArray* rest = [nameArray subarrayWithRange:r];
                if([self isNamingContext:obj]){
                    return [obj resolve:rest];
                }else{
                    CORBAException* excp = [_interface exceptionWithName:@"CannotProceed"];
                    [excp takeValue:obj forKey:@"ctx"];
                    [excp takeValue:rest forKey:@"rest_of_name"];
                    [excp raise];
                }
            }
        }else{
            CORBAException* excp = [_interface exceptionWithName:@"NotFound"];
            id reason = [_interface enumValueForName:@"NotFoundReason" andKey:@"missing_node"];
            [excp takeValue:reason forKey:@"why"];
            [excp takeValue:nameArray forKey:@"rest_of_name"];
            [excp raise]; 
        }
    }else{
        CORBAException* excp = [_interface exceptionWithName:@"InvalidName"];
        [excp raise];
    }
    return nil;
}

- (void) unbind:nameArray
{
    int cnt = [nameArray count];
    if(cnt > 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id obj = [_entries objectForKey:key];
        if(obj  && [self isNamingContext:obj]){
            NSRange r = {1,cnt-1};
            NSArray* rest = [nameArray subarrayWithRange:r];
            [obj unbind:rest];
            return;
        }
    }else
    if(cnt == 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id obj = [_entries objectForKey:key];
        if(![self isNamingContext:obj]){
            [_entries removeObjectForKey:key];
            return;
        }
    }
    [[_interface exceptionWithName:@"InvalidName"] raise];
}

- newContext
{
    return [[[NamingContext alloc]init]autorelease];
}

- bindNewContext:nameArray
{
    int cnt = [nameArray count];
    if(cnt > 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id obj = [_entries objectForKey:key];
        if(obj  && [self isNamingContext:obj]){
            NSRange r = {1,cnt-1};
            NSArray* rest = [nameArray subarrayWithRange:r];
            return [obj bindNewContext:rest];
        }
    }else
    if(cnt == 1){
        NSString* key = [self stringFromName:[nameArray objectAtIndex:0]];
        id obj = [_entries objectForKey:key];
        if(obj)
            [[_interface exceptionWithName:@"AlreadyBound"] raise];
        else{
            obj = [self newContext];
            [_entries setObject:obj forKey:key];
            return obj;
        }
    }
    [[_interface exceptionWithName:@"InvalidName"] raise];
    return nil;
}

- (void) destroy
{
    // not implemented
}

- (void) list:(unsigned long) count bindingList:(id*)list bindingIterator:(id*)iterator
{
    // Ignore 'count' and return all the entries
    NSMutableArray* array = [NSMutableArray array];
    NSEnumerator* en = [_entries keyEnumerator];
    NSString* key;
    while((key = [en nextObject]) && (count-- > 0)){
        id obj = [_entries objectForKey:key];
        id type;
        NSDictionary* dict;
        if([self isNamingContext:obj])
            type = [_interface enumValueForName:@"BindingType" andKey:@"ncontext"];
        else
            type = [_interface enumValueForName:@"BindingType" andKey:@"nobject"];
        dict = [NSDictionary dictionaryWithObjectsAndKeys:
            [self nameFromString:key], @"binding_name",
            type, @"binding_type", nil];
        [array addObject:dict];
    }
    *list = array;
    *iterator = nil;
}

+ (void) runNameService
{
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
    NamingContext* ctx = [[[NamingContext alloc]init]autorelease];
    CORBAObject* parentNaming = [ORB resolveInitialReferences:@"NameService"];
    NSString* path = [[NSUserDefaults standardUserDefaults]stringForKey:@"bind"];
    NSString* iorf = [[NSUserDefaults standardUserDefaults]stringForKey:@"iorf"];

    [ctx activateWithObjectKey:@"NameService"];
    
    // If both 'parentNaming' and 'path' are defined then rebind ourselves in 'parentNaming'
    if(parentNaming && path){
        NS_DURING
        [parentNaming rebindContext:[IDLNameComponent nameComponentsForPath:path] namingContext:ctx];
        NS_HANDLER
        NSLog(@"%@: %@, userInfo: %@\n",
            [localException name], [localException reason], [[localException userInfo]description]);
        NS_ENDHANDLER
    }
    if(iorf){ // save IOR to file
        if([iorf characterAtIndex:0] != '/')
        	iorf = [[[NSFileManager defaultManager]currentDirectoryPath]
                        stringByAppendingPathComponent:iorf];
        [[[ctx objectReference] toString]writeToFile:iorf atomically:NO];
    }
    
    [pool release];
    [ORB run];
}

@end

int main (int argc, const char *argv[])
{
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    [NamingContext runNameService];
    [pool release];
    exit(0);       // insure the process exit status is 0
    return 0;      // ...and make main fit the ANSI spec.
}

