/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDL.h"
#import "IDLTCKind.h"
#import <expr.tab.h>

@interface IDLDefinition(Header)
- (NSString*) argumentType;
@end

@interface IDLFunction(Header)
- (NSString*) headerString:(NSString*)selName;
- (NSArray*) interfaceReferences;
@end

@implementation IDLInterface(Header)

static NSString* RootProtocol = @"CORBAObject";

- (NSString*)protocolName
{
	NSString* str = [self name];
	if([str isEqualToString:@"Object"])
		return RootProtocol;
	else
		return str;
}

- (NSString*)generateHeader
{
    NSMutableString* header = [NSMutableString stringWithFormat:
        @"\n// Generated by ADORB, %@\n\n#import <Foundation/Foundation.h>\n\n"
		@"#import <ADORB/CORBAObject.h>\n\n",
            [NSDate date]];
    NSAutoreleasePool* pool0 = [[NSAutoreleasePool alloc]init];
    NSMutableString* body = [NSMutableString stringWithFormat:@"@protocol %@",[self protocolName]];
    NSString* tail = [NSString stringWithFormat:
        @"\n@interface CORBAObject(%@) <%@>\n\n@end\n", [self name], [self protocolName]];
    NSMutableArray* references = [NSMutableArray arrayWithObject:RootProtocol];
    
    NSEnumerator* en = [_inheritance objectEnumerator];
    IDLInterface* obj;
    IDLFunction* function;
    
    while(obj = [en nextObject])
        [references addObject:[obj protocolName]];
    
    if([references count] > 0)
        [body appendFormat:@"<%@>\n", [references componentsJoinedByString:@","]];
    else
        [body appendString:@"\n"];
    
    en = [[self allFunctions]objectEnumerator];
    while(function = [en nextObject]){
        NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
        NSArray* array = [function interfaceReferences];
		NSString* sel = [function selectorName];
        [body appendFormat:@"%@;\n",[function headerString:sel]];
        if([array count] > 0){
            NSEnumerator* en2 = [array objectEnumerator];
            while(obj = [en2 nextObject]){
                NSString* objName = [obj name];
                if(![references containsObject:objName] &&
                        ![objName isEqualToString:[self name]])
                    [references addObject:objName];
            }
        }
        [pool release];
    }

    [body appendString:@"@end\n\n"];
    if([references count] > 0){
        NSString* str;
        en = [references objectEnumerator];
        while(str = [en nextObject])
            [header appendFormat:@"@protocol %@;\n",str];
    }
    [header appendFormat:@"\n%@\n\n%@\n", body, tail];
    [pool0 release];

    return header;
}

- (NSString*) argumentType
{
    return [NSString stringWithFormat:@"id<%@>",[self protocolName]];
}

@end

@implementation IDLFunction(Header)

- (NSString*) headerString:(NSString*)selName
{
    NSMutableString* header = [NSMutableString string];
    
    NSString* t = [[self returnType]argumentType];
    if([t length] > 0)
        [header appendFormat:@"- (%@)",t];
    else
        [header appendString:@"-"];
    
    if([_args count] > 0){
        NSArray* array = [selName componentsSeparatedByString:@":"];
        int n = 0;
        for(n=0;n<[_args count];n++){
            IDLArgument* def = [_args objectAtIndex:n];
            NSString* ident = [def name];
            BOOL isPtr = [def isOutMode];
            t = [[def definition]argumentType];
            
            [header appendFormat:@" %@:",[array objectAtIndex:n]];
            if(isPtr){
                if([t length] == 0)
                    t = @"id";
                [header appendFormat:@"(%@*)",t];
            }else{
                if([t length] > 0)
                    [header appendFormat:@"(%@)",t];
            }
            [header appendString:ident];
        }
        if(_contextList)
            [header appendString:@" context:context"];
    }else{
        [header appendFormat:@" %@", [self selectorName]];
        if(_contextList)
            [header appendString:@"context"];
    }
    return header;
}

- (NSArray*) interfaceReferences
{
    NSMutableArray* array = [NSMutableArray array];
    NSEnumerator* en = [_args objectEnumerator];
    IDLArgument* def;
    IDLDefinition* obj;
    while(def = [en nextObject]){
        obj = [def definition];
        if([obj isKindOfClass:[IDLInterface class]])
            [array addObject:obj];
    }
    obj = [self returnType];
    if([obj isKindOfClass:[IDLInterface class]])
        [array addObject:obj];
    
    if([array count] > 0)
        return array;
    else
        return nil;
}

@end

@implementation IDLDefinition(Header)
- (NSString*) argumentType
{
    return nil;
}
@end
@implementation IDLOctetArray(Header)
- (NSString*) argumentType
{
    return @"NSData*";
}
@end
@implementation IDLEnum(Header)
- (NSString*) argumentType
{
    return @"long";
}
@end
@implementation IDLSequence(Header)
- (NSString*) argumentType
{
	if([type kind] == tk_octet)
		return @"NSData*";
	else
		return @"NSArray*";
}
@end
@implementation IDLOctetSequence(Header)
- (NSString*) argumentType
{
	return @"NSData*";
}
@end
@implementation IDLArray(Header)
- (NSString*) argumentType
{
	if([_type kind] == tk_octet)
		return @"NSData*";
	else
		return @"NSArray*";
}
@end
@implementation IDLPrimitive(Header)
- (NSString*) argumentType
{
    switch(_type){
        case CHAR:
            return @"char";
        case WCHAR:
            return @"unsigned short";
        case OCTET:
			if([IDLPrimitive mapsOctetToNSData])
				return @"NSData*";
			else
				return @"unsigned char";
        case SHORT:
            return @"short";
        case USHORT:
            return @"unsigned short";
        case INT:
            return @"int";
        case UINT:
            return @"unsigned int";
        case LONG:
            return @"long";
        case ULONG:
            return @"unsigned long";
        case UNSIGNED:
            return @"unsigned";
        case LONGLONG:
            return @"long long";
        case ULONGLONG:
            return @"unsigned long long";

        case FLOAT:
            return @"float";
        case DOUBLE:
            return @"double";
        case FIXED:
            return @"NSNumber*";
        case BOOLEAN:
            return @"BOOL";
        case STRING:
        case WSTRING:
            return @"NSString*";

        case VOID:
            return @"void";
        case ONEWAY:
            return @"oneway void";
        default:
            return nil;
    }
}
@end
@implementation IDLStruct(Header)
- (NSString*) argumentType
{
	NSString *className = [self objectClassName];
	if(className)
		return [className stringByAppendingString:@"*"];
	else
		return @"NSDictionary*";
}
@end
@implementation IDLUnion(Header)
- (NSString*) argumentType
{
	NSString *className = [self objectClassName];
	if(className)
		return [className stringByAppendingString:@"*"];
	else
		return @"NSDictionary*";
}
@end
