/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "IDLDefinition_Private.h"
#import "IDLValueType_Private.h"
#import "IDLValueClass.h"
#import "IDLDefinitionScope.h"
#import "IDLEncapsulation.h"
#import "IDLLink.h"
#import "IDLValueObject.h"

@implementation IDLValueType(CORBACoding)

static BOOL logEnabled = NO;

+ (void) setLogEnabled:(BOOL)flag;
{
    logEnabled = flag?YES:NO;
}

- (void) encodeObject:obj withCoder:(id <CORBACoder>)coder
{
    IDLValueType* type = [self valuetypeForObject:obj];
    [coder encodeValue:obj forObjectCoder:type];
}

- (void) decodeObject:(id*)obj withCoder:(id <CORBADecoder>)coder
{
	int pos0 = 0;
	if(logEnabled)
		pos0 = [coder cursorPosition];
    [coder decodeValue:obj forObjectCoder:self];
	if(logEnabled)
		NSLog(@"valuetype %@: decoded value 0x%x range %d.%d", _name, *obj, pos0, [coder cursorPosition] - pos0);
}

// This method is called after obj is decoded, to finalize or replace obj
// If self is ValueBase then objCoder is different, but it may be not passed anywhere
// In this case we wrap it into IDLValueObject
- (void) finishDecodingObject:(id*)obj withObjectCoder:(IDLValueType*)objCoder
{
	// if self is ValueBase and *obj was not created with initWithIDLType 
	// and could not be set by setIDLType:
	// Then create value wrapper
	if(!_support && *obj && 
			![*obj respondsToSelector:@selector(initWithIDLType:)] &&
			![*obj respondsToSelector:@selector(setIDLType:)]) {
		IDLValueObject* newObject = [IDLValueObject objectWithIDLType:objCoder];
		[newObject setWrappedValue:*obj];
		*obj = newObject;
	}
}
- (void) decodeObjectState:(id*)obj withCoder:(id <CORBADecoder>)coder
{
    if(logEnabled)
        NSLog(@"valuetype %@: decoding state for 0x%x at %d\n", _name, *obj, [coder cursorPosition]);
    if([_support respondsToSelector:@selector(decodeObjectState:withCoder:)])
        [_support decodeObjectState:obj withCoder:coder];
    else
        [_support decodeObject:obj withCoder:coder];
    
	// If *obj was not initialized with -initWithIDLType:
	// then try to set IDL type using -setIDLType: method
    if(![*obj respondsToSelector:@selector(initWithIDLType:)] &&
			[*obj respondsToSelector:@selector(setIDLType:)])
        [*obj setIDLType:self];
}

- (void) encodeObjectState:obj withCoder:(id<CORBACoder>)coder
{
    if(logEnabled)
        NSLog(@"valuetype %@: encoding value for 0x%x\n", _name, *obj);
    if([_support respondsToSelector:@selector(encodeObjectState:withCoder:)])
        [_support encodeObjectState:obj withCoder:coder];
    else
        [_support encodeObject:obj withCoder:coder];
}

- (const char*) objCType
{
    return @encode(id);
}
- (unsigned) argLength
{
    return sizeof(id);
}
- (unsigned) returnValueLength
{
    return sizeof(id);
}

- argumentObjectFromInvocation:(NSInvocation*)inv atIndex:(int)n
{
    // IDLValueType is mapped to NSObject
    id object = nil;
    [inv getArgument:&object atIndex:n];
    return object;
}

- argumentObjectOfType:(const char*)argType fromPointer:(void*)ptr
{
    id object = *((id*)ptr);
    return object;
}

- (void) setValue:value ofType:(const char*)argType forPointer:(void*)ptr
{
    *((id*)ptr) = value;
}
- (void) setValue:value forInvocation:(NSInvocation*)inv atIndex:(int)n
{
    id _val = value;
    [inv setArgument:&_val atIndex:n];
}
- (void) setReturnValue:value forInvocation:(NSInvocation*)inv
{
    id _val = value;
    [inv setReturnValue:&_val];
}
- returnValueFromInvocation:(NSInvocation*)inv
{
    id _val = nil;
    [inv getReturnValue:&_val];
    return _val;
}

- (void) _encodeWithCORBACoder:(id <CORBACoder>)coder
{
    [coder encodeString:[self typeId]];
    [coder encodeString:[self name]];
    if(_support)
        [_support encodeWithCORBACoder:coder];
    else
        [coder encodeLong:tk_null];
}

- (void) encodeWithCORBACoder:(id <CORBACoder>)coder
{
    long location = [coder locationForDefinition:self];
    if(location != 0 && [IDLDefinition allowsIndirection])
        [coder encodeIndirection:location];
    else{
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [coder encodeLong:[self kind]];
        [stream setParentStream:coder location:[coder location]];
        [stream registerLocation:-2*sizeof(long) forDefinition:self];
        [self _encodeWithCORBACoder:stream];
        [coder encodeData:[stream data]];
    }
}


+ _definitionFromCoder:(IDLEncapsulation*)coder tag:(long)tag
{
    IDLLink* link = [IDLLink linkWithDefinition:nil];
    NSString *str;
    NSString* typeId;
    IDLValueType* obj = nil;
    IDLValueType* existingObj = nil;

    [coder registerDefinition:(IDLDefinition*)link forLocation:-2*sizeof(long)];

    [coder decodeString:&typeId];
    [coder decodeString:&str];

    obj = [IDLValueType valueTypeWithName:str];
    [obj setTypeId:typeId];
    if(tag == tk_value_box){
        id support = [IDLDefinition definitionFromCoder:coder];
        [obj setTypeDefinition:support];
    }else{
        [[obj classSupport]decodeFromCoder:coder];
        [obj setModifier:[[obj classSupport] modifier]];
    }

    existingObj = (IDLValueType*)[[IDLDefinitionScope currentScope]definitionForTypeId:typeId];
    if([existingObj isEqual:obj])
        obj = existingObj;

    if(obj){
        [link setLink:obj];
        return link;
    }else
        [NSException raise:CORBAMarshalException 
            format:@"unable to un-marshal TypeCode, at %d", [coder location]];
    return nil;
}

+ _definitionFromCoder:(IDLEncapsulation*)coder
{
    [NSException raise:CORBANotImplemented format:@"demarshaling typecode for valuetype is not implemented"];
    return nil;
}

@end

